/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1984, 2024 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// For updating from 14.04 to 14.05.

#include "eel.h"
#include "c.h"
#include "colcode.h"
#include "diff.h"
#include "dired.h"
#include "html.h"
#include "perl.h"
#include "python.h"
#include "session.h"

// Move after the pair of {} or [] or () starting just after pos, and
// return 1 if found, else return 0 and don't move.
int c_fwd_skip_delims(int pos)
{
	int ret = 0, orig = point;
	point = pos;
	switch (character(pos)) {
		case '{':	ret = c_skip_level('{', '}'); break;
		case '[':	ret = c_skip_level('[', ']'); break;
		case '(':	ret = c_skip_level('(', ')'); break;
	}		
	if (!ret)
		point = orig;
	return ret;
}

full_comp_read(trystr, prmpt, finder, flags, edit, def)
char *trystr, *prmpt, *def;
char *(*finder)();
{
	int i, j, num_match = 0, maxlen = ptrlen(trystr) - 1, pbuf, res;
	int pwin, editwin, owin = window_handle, recall_cnt = 0;
	int olen = 0, in_menu = 0, menu = -1, mbuf = 0;
	jmp_buf this_level;
	int *(*lister)();
	char password_display_class[256];

	save_var last_index;
	save_var _help_on_cmd = last_index;
	save_var _doing_input = DI_LINEINPUT;
	save_var _doing_completion = (finder != 0);

	while (!in_macro() && !char_avail() && unseen_msgs())
		check_abort();	// Wait for unseen msgs to display so we don't cover them.
	pwin = add_popup(0, echo_line, screen_cols - 1, 1, 0,
		pbuf = tmp_buf());	/* make system window with prompt */
	stuff(prmpt);
	i = current_column();
	system_window = 1;	/* don't let user switch to it */
	point = display_column = 0;
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;

	editwin = add_popup(i, echo_line, screen_cols - i, 1,
							0, edit);
	display_column = 0;	/* horizontally scroll this window */
	text_color = color_class echo_area;
	_display_class = _echo_disp_class;
	if (flags & PASSWORD_PROMPT) {
		set_chars(password_display_class, '*', 256);
		_display_class = password_display_class;
		build_first = 1;
		save_var disable_recording_last_keys = 1;
	}
	mouse_handler = allow_mouse_switching;
	if (insert_default_response == 1) {
		highlight_on();
		shift_selecting = 2;
	}
	if ((flags & POP_UP_PROMPT) && (has_feature & FEAT_GUI_DIALOGS)) {
		remove_window(pwin);
		do_display_dialog_box("OneLineBox", prmpt, editwin, 0, 0,
							  "OK", "Cancel", "Help", NULL, NULL);
	}
	maybe_refresh();
	save_var top_level = &this_level;
	if (setjmp(top_level)) {
		if (in_menu == 2) {	/* menu is half-created */
			remove_window(menu);
			buf_delete(mbuf);
			in_menu = mbuf = 0;
			if (is_gui && want_gui_prompts) {
				res = INP_ABORT;
				goto done;
			}
		}
	}
	for (;;) {
		if (mbuf) {
			window_handle = menu;
			if (in_menu && line_gets_highlight())
				make_line_highlight();
			else
				remove_line_highlight();
		}
		window_handle = in_menu ? menu : editwin;
		mode_keys = in_menu ? menu_tab : comp_tab;
		_doing_input = (response_tracks_list && mbuf)
					   ? DI_LISTBOX : DI_LINEINPUT;
		if (_doing_input == DI_LISTBOX) {
			maybe_refresh();
			getkey();
			cmd_result = 0;
			run_topkey();
			res = cmd_result;
		} else
			res = do_recursion();	/* let user edit input */
		if (!is_window(editwin) || in_menu && !is_window(menu)) {
			res = INP_ABORT;
			break;
		} else if (window_handle == (in_menu ? editwin : menu))
			in_menu = !in_menu;
		window_handle = editwin;
		if (res == INP_ADD) {
			window_handle = editwin;
			run_topkey();
			if (search_in_menu && mbuf) {
				window_handle = editwin;
				grab(0, maxlen, trystr);
				if (strlen(trystr) >= olen) {
					window_handle = menu;
					point = 0;
					col_search(trystr, 0);
				}
			} else
				in_menu = 0;
		} else if (res == INP_SEEN)
			in_menu = 0;
		else if (res == INP_TO_LIST && mbuf) {
			window_handle = menu;
			in_menu = 1;
		} else if ((res == INP_GET || res == INP_GET_EXIT) && mbuf) {
			window_handle = menu;
			get_menu_line(edit);
			window_handle = editwin;
			in_menu = 0;
		} else if (!res) {
			if (_doing_input == DI_LISTBOX)
				if (in_menu) {
					window_handle = menu;
					get_menu_line(edit);
				} else {
					window_handle = editwin;
					grab(0, maxlen, trystr);
					window_handle = menu;
					point = 0;
					col_search(trystr, 0);
				}
			continue;
		}
		restart_file_paths();
		if (res == INP_ABORT || res == INP_GET_EXIT || res == INP_EXIT)
			break;
		else if (res == INP_GRAB_WORD) {
			grab_buffer_word(owin, 0);
			continue;
		} else if (res == INP_GRAB_LINE) {
			grab_buffer_word(owin, 1);
			continue;
		} else if (res == INP_DEFAULT) {
			if (!size())	/* type in default text */
				stuff(def);
			in_menu = 0;
			continue;
		} else if (res == INP_LAST_RECALL || res == INP_NEXT_RECALL
				   || res == INP_PREV_RECALL) {
			if (res == INP_LAST_RECALL)
				get_last_recall(finder, trystr);
			else
				recall_cnt = get_last_recall_from_end(finder, trystr,
					recall_cnt + (res == INP_NEXT_RECALL ? -1 : 1));
			if ((*trystr || res != INP_LAST_RECALL) && !mbuf) {
				delete(0, size());
				stuff(trystr);
				if (recall_response_selected && *trystr) {
					highlight_on();
					shift_selecting = 2;
				}
			}
			in_menu = 0;
		} else if (res == INP_GET_OTHER) {
			if (inp_get_other_name && *inp_get_other_name) {
				delete(0, size());
				stuff(inp_get_other_name);
			}
			in_menu = 0;
		} else if (res == INP_SHOW_MATCHES || res == INP_RECALL) {
show_matches:		grab(0, maxlen, trystr);	/* typed '?' */
			if (is_highlight_on() || res == INP_RECALL) {
				if (flags & COMP_FILE_BIT)
					*get_tail(trystr, 0) = 0;
				else
					*trystr = 0;
				highlight_off();
			}
			olen = strlen(trystr);
			if (!mbuf) {
				menu = new_menu(mbuf = tmp_buf());
				if (is_gui && want_gui_prompts) {
					remove_window(pwin);
					say("");
					make_prompt_box(res == INP_RECALL, prmpt, editwin, menu, finder);
				}
			}
			in_menu = 2;		/* initializing */
			if (res == INP_RECALL)
				lister = list_recalls;
			else if (completion_lister)
				lister = completion_lister;
			else
				lister = list_matches;
			make_menu(menu, trystr, finder, flags, editwin, mbuf, lister);
			in_menu = 1;
			mode_keys = menu_tab;
			cursor_to_column = -1;
		} else if (res == INP_TRY || res == INP_COMPLETE) {
			if (res == INP_TRY && cursor_to_column
						== current_column() + 1)
				break;	/* exactly 1 match last time */
			grab(0, maxlen, trystr);
			i = strlen(trystr);
			num_match = complete(trystr, finder, flags);
			j = strlen(trystr);
			if ((flags & SPACE_VALID) && key == ' ' && trystr[i] == ' '
				  && could_add_space(trystr, finder, flags) == 1)
				i++, num_match++;
			else if (num_match && i != j)
				i = j;
			else if (num_match != 1) {
				if ((flags & SPACE_VALID) && key == ' ' && is_window(menu)
//					  && trystr[i] == ' '
// Given dirs c:\program files and c:\programs, this commented-out code
// made selecting the former fail.
					  && could_add_space(trystr, finder, flags)) {
					trystr[i++] = ' ';
				} else {
					maybe_ding(bell_on_completion);
					if (completion_pops_up)
						ungot_key = '?';
				}
			}
			if ((flags & COMP_FILE_BIT) && num_match == 1 && i > 0
			    && is_path_separator(trystr[i - 1]))
				num_match++;	/* might be directory */
			trystr[i] = 0;
			buf_zap(edit);
			stuff(trystr);
			if (num_match == 1) {
				if (res == INP_COMPLETE)
					break;	/* <Esc> worked */
				maybe_refresh();
				cursor_to_column = current_column() + 1;
			}
		}
	}
done:
	if (res != INP_ABORT) {
		grab(0, maxlen, trystr);
		if (size() > maxlen)
			say("Response exceeds internal limit of %d characters; truncated.",
				  maxlen);
	} else if (!user_abort)		/* user deleted edit window */
		user_abort = 1;
	if (res == INP_EXIT && must_match(trystr, finder, flags))
		goto show_matches;
	if (is_window(owin))
		window_handle = owin;
	remove_window(editwin);
	remove_window(pwin);
	buf_delete(pbuf);
	if (mbuf) {
		remove_window(menu);
		buf_delete(mbuf);
	}
}

char dired_dirsearch_unixfull[] = "^. [dla]";

dired_isdir_unixfull()
{
	return dired_isdir_unix();
}

int is_unc_path(char *fname)	// A UNC path like \\server\share\path\file?
{
	return is_path_separator(fname[0]) == 1 && is_path_separator(fname[1]) == 1;
}

xwin_set_font_first_prune()
{
	point = 0;
	// Now, discard proportional fonts.
	do_drop_matching_lines(REGEX, "-p-[^-\n]*-[^-\n]*-[^-\n]*$", 1);
	// Discard font aliases too.
	do_drop_matching_lines(REGEX, "^-[^-\n]*-[^-\n]*-[^-\n]*-[^-\n]*-[^-\n]*-", 0);
	// Also drop any fonts with weird charsets.
	do_drop_matching_lines(REGEX, "-fontspecific(-[^-\n]*)?$", 1);
	// XQuartz 2.8.2-2.8.4 includes these fonts that cause any app selecting
	// them to crash on M1 (works on Intel). Don't use them.
	do_drop_matching_lines(REGEX, "^-misc-gb18030 bitmap-thin-.*$", 1);
}

int is_php_alt_outdent(int p)
{
	if (sublanguage_code != SUBLANG_PHP)
		return 0;
	save_var point = p;
	if (parse_string(1, "(end(if|while|for|foreach|switch))</word>|else:"))
		return 1;
	if (parse_string(1, "elseif<wspace>*%(")) {
		point = matchend - 1;
		if (c_fwd_skip_delims(point) && parse_string(1, "<wspace>*:"))
			return 1;
	}
	return 0;
}

rmode_paren_follows() // is there whitespace followed by a "(" at point in buffer?
{		// if so, we assume it's a function name
	save_var point;
	point += parse_string(1, "[ \t\n]*");
	for (;; point++) {
		switch (curchar()) {
			case ' ': case '\t': case '\n':
				continue;
			case '(':		return 1;
			case '#':		nl_forward();
							break;
			default:		return 0;
		}
	}
}

remember_this_buffer()	/* should we record this buffer in session file? */
{
	char *fname = filename;
	int limit;

	if (*bufname == '-')
		return 0;
	if (is_process_buffer(bufnum) || (buffer_not_saveable & (NOSAVE_ALL | NOSAVE_SESSION))
		  || _read_aborted || (force_save_as & FSA_NEWFILE))
		return 0;
	if (file_info.year && file_info.check_type != CHECK_FILE)
		return 0;
	if ((session_restore_exclude_file_types & 8) && is_unc_path(fname))
		return 0;
	if (is_dired_buf()) {
		fname = get_dired_buffer_pattern();
		if (!(session_restore_directory_buffers
			  & (is_remote_file(fname) ? 2 : strstr(fname, "**") ? 4 : 1)))
			return 0;
	} else {
		if (!*filename)
			return 0;
		limit = is_remote_file(filename) ? session_restore_biggest_remote_file
										 : session_restore_biggest_file;
		if (limit >= 0 && size() > limit)
			return 0;		// too big to restore
		if (!size() && check_file(filename) != CHECK_FILE)
			return 0;
	}
	if (session_restore_exclude_file_types) {
		int flags = filename_rules(fname);
		if ((flags & FSYS_NETWORK) && (session_restore_exclude_file_types & 1))
			return 0;
		if ((flags & FSYS_REMOVABLE) && (session_restore_exclude_file_types & 2))
			return 0;
		if ((flags & FSYS_CDROM) && (session_restore_exclude_file_types & 4))
			return 0;
	}
	return 1;
}
